@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2000 Michal Trojnara <Michal.Trojnara@centertel.pl>
+ *   Copyright (c) 1998-2000 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -27,38 +27,46 @@
 #define OPT_TRANSPARENT 0x40
 #define OPT_PTY         0x80
 
-#ifdef USE_WIN32
+/* Certificate defaults */
 
-/* default certificate */
-#define DEFAULT_CERT "stunnel.pem"
+/* let's not use openssl defaults unless told to at compile time. */
+#ifndef SSLLIB_CS
+#define SSLLIB_CS	0
+#endif
 
-/* additional directory (hashed!) with trusted CA client certs */
-#define CA_DIR "trusted"
+#define SSL_CERT_DEFAULTS	1
+#define STUNNEL_CERT_DEFAULTS	2
 
-/* certificate used for sign our client certs */
-#define CLIENT_CA "cacert.pem"
+#define CERT_DEFAULTS ( SSLLIB_CS | STUNNEL_CERT_DEFAULTS )
 
-#else /* USE_WIN32 */
 
-/* default certificate */
-#define DEFAULT_CERT certdir "/stunnel.pem"
+/* Set some defaults if not set in makefiles */
+#ifndef CERT_DIR
+#define CERT_DIR	""
+#endif
+#ifndef CERT_FILE
+#define CERT_FILE	""
+#endif
+#ifndef PEM_DIR
+#define PEM_DIR	""
+#endif
 
-/* additional directory (hashed!) with trusted CA client certs */
-#define CA_DIR certdir "/trusted"
-
-/* certificate used for signing our client certs */
-#define CLIENT_CA ssldir "/localCA/cacert.pem"
-
-#endif /* USE_WIN32 */
 
 #ifdef USE_WIN32
 
-#define VERSION "3.8"
-#define HOST "i586-pc-mingw32-gnu"
+#define VERSION "3.10"
+#ifdef __MINGW32__
+#define HOST "x86-pc-mingw32-gnu"
+#else
+#define HOST "x86-pc-unknown"
+#endif
+
 
 #define HAVE_VSNPRINTF
 #define vsnprintf _vsnprintf
+/*  Already defined for mingw, perhaps others
 int _vsnprintf(char *, int, char *, ...);
+*/
 
 #define get_last_socket_error() WSAGetLastError()
 #define get_last_error()        GetLastError()
@@ -126,6 +134,10 @@ extern char *sys_errlist[];
 /* Length of strings (including the terminating '\0' character) */
 #define STRLEN         1024
 
+/* How many bytes of random input to read from files for PRNG */
+/* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
+#define RANDOM_BYTES		64
+
 /* Safe copy for strings declarated as char[STRLEN] */
 #define safecopy(dst, src) \
     (dst[STRLEN-1]='\0', strncpy((dst), (src), STRLEN-1))
@@ -133,8 +145,9 @@ extern char *sys_errlist[];
     (dst[STRLEN-1]='\0', strncat((dst), (src), STRLEN-strlen(dst)-1))
 
 typedef struct {
-    char certfile[STRLEN];  /* name of the certificate */
-    char clientdir[STRLEN];
+    char pem[STRLEN];  		/* pem (priv key/cert) filename */
+    char cert_dir[STRLEN];	/* directory for hashed certs */
+    char cert_file[STRLEN];	/* file containing bunches of certs */
     char pidfile[STRLEN];
     unsigned long dpid;
     int clients;
@@ -146,13 +159,20 @@ typedef struct {
     char servname[STRLEN];  /* service name for loggin & permission checking */
     int verify_level;
     int verify_use_only_my;
-    int debug_level;
+    int debug_level;		/* debug level for syslog */
+    int facility;		/* debug facility for syslog */
     long session_timeout;
     char *cipher_list;
     char *username;
     char *protocol;
     char *setuid_user;
     char *setgid_group;
+    char *egd_sock;	/* entropy gathering daemon socket */
+    char *rand_file;	/* file with random data */
+    int rand_write;	/* overwrite rand_file with new rand data when PRNG seeded */
+    int random_bytes;	/* how many random bytes to read */
+    char *pid_dir;
+    int cert_defaults;
 } server_options;
 
 /* Prototypes for stunnel.c */
@@ -176,9 +196,12 @@ int negotiate(char *, int, int, int);
 void log_open();
 void log_close();
 void log(int, char *, ...);
+int  parse_debug_level(char *);
 
 /* Prototypes for sthreads.c */
 
+void enter_critical_section(int);
+void leave_critical_section(int);
 void sthreads_init(void);
 unsigned long process_id(void);
 unsigned long thread_id(void);
@@ -191,5 +214,14 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *ttyname, int ttynamelen);
 void pty_release(char *ttyname);
 void pty_make_controlling_tty(int *ttyfd, char *ttyname);
 
+/* define for windows, although ignored */
+#ifndef PIDDIR
+#define PIDDIR ""
+#endif
+
+#define STRINGIFY_H(x) #x
+#define STRINGIFY(x) STRINGIFY_H(x)
+
+
 /* End of common.h */
 
